/* #Specification: aclfs / .acl file format
	The aclfsd daemon relies on ..acl files distributed in
	directories. The file is an ascii file for now. It may change
	at some point to a binary file to make things a little
	faster, but aclfsd is doing heavy caching (and must be noticed)
	to reload a file.

	Here is a sample

	#
	list list_name {
		jack, @users
	};
	list otherlist {
		# list may reference other lists
		joe,&list_name
	};

	file pattern {
		context name{
			grant { &nom }
			deny {	&nom2 }
			rights read,overwrite,append;
		}
		context name2{
			programs /bin/cat
			module /modulepath args
			grant{
				&nom,jack,@users
			}
			deny {
				&nom2
			}
			rights read,overwrite,append;
		}
	};
	#

	Note that the syntax is free formed. A definition may be splitted
	in several lines for clarity.

	The "list" keyword let you define list of item. An item may be
	a user, a group or another list. User are identified using their
	name, group are prefixed with the @ character and list are
	prefixed by the & character.

	Lists may be defined in the ..acl files as well as in the main
	/etc/aclfsd.conf (to be added one day).

	You can define several lists.

	Then you define privileges for the various entries in the
	directory. You can use regexp to define file pattern. The
	pattern * represent all files.

	Note that the ..acl is processed sequentially, so the first
	file pattern that match is the only one that will be used
	for a file operation. Note that the file section
	handles both files and sub-directories.

	Within a file section, you have several context. Each context
	associate grant and deny lists to few privileges. The processing
	is done this way:

	<SGML>
	<itemize>
	<item>We start a query by resetting all privileges.
	<item>We walk one context after the other. If the user is
		a member of the grant list (either the user or the
		group it is a member of), then we add all the
		privileges defined in the access line. If the user
		is a member of the deny list, then we reset all privileges
		in the access line.
	<item>We continue to the next context.
	</itemize>
	</SGML>

	A context may define a list of program. The context will apply
	only if the user is running one of those programs.

	A context may also reference an aclfsd module (to be invented
	at some point). The module will be passed all the conext and
	will perform ... something. Modules will be able to
	tell if a context apply, or may modify the access rights.
	A module may even define how the file is open and provide
	a cooked solution (the calling program will read and write to
	a wirtual file...).

*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <syslog.h>
#include <grp.h>
#include <pwd.h>
#include "aclfsd.h"

PUBLIC ACL_ITEM::ACL_ITEM (ACLTOK &tok, const ACL_LISTS &lists)
{
	next = NULL;
	list = NULL;
	id = -1;
	type = ITEM_UID;
	const char *s = tok.token;
	if (tok.type == ACLTOK_GROUP){
		struct group *g = getgrnam(s);
		if (g==NULL){
			tok.logwarning ("Unknown group %s",s);
		}else{
			id = g->gr_gid;
			type = ITEM_GID;
		}
	}else if (tok.type == ACLTOK_LIST){
		const ACL_LIST *l = lists.first;
		while (l != NULL){
			if (strcmp(l->name,s)==0){
				list = l;
				break;
			}
			l = l->next;
		}
		if (list == NULL){
			tok.logwarning ("Unknown list %s",s);
		}
	}else if (tok.type == ACLTOK_ID){
		if (tok.cmp("ALL")==0){
			type = ITEM_ALL;
		}else{
			struct passwd *p = getpwnam(s);
			if (p == NULL){
				tok.logwarning ("Unknown user %s",s);
			}else{
				id = p->pw_uid;
				type = ITEM_UID;
			}
		}
	}
}

PUBLIC ACL_LIST::ACL_LIST()
{
	next = NULL;
	first = NULL;
}

/*
	Parse and record a list, starting with the {
*/
PUBLIC void ACL_LIST::parse(
	ACLTOK &tok,
	ACL_LISTS &lists)
{
	if (tok.next() == ACLTOK_OPNBRAC){
		bool expectid = true;
		while (tok.next() != ACLTOK_EOF){
			if (tok.type == ACLTOK_CLSBRAC){
				break;
			}else if (expectid){
				if (tok.type == ACLTOK_ID
					|| tok.type == ACLTOK_GROUP
					|| tok.type == ACLTOK_LIST){
					add (new ACL_ITEM(tok,lists));
				}else{
					tok.logerror ("Syntax error (expect id, group or list), got %s"
						,tok.token);
				}
				expectid = false;
			}else if (tok.type != ACLTOK_COMMA){
				tok.logsyntax ();
			}else{
				expectid = true;
			}
		}
	}else{
		tok.logsyntax();
	}
}

PUBLIC ACL_LIST::ACL_LIST (
	ACLTOK &tok,
	ACL_LISTS &lists)
{
	next = NULL;
	first = NULL;
	if (tok.next() == ACLTOK_ID){
		name = strdup (tok.token);
		parse (tok,lists);
	}else{
		tok.logsyntax();
	}
}

PUBLIC ACL_LIST::~ACL_LIST()
{
	while (first != NULL){
		ACL_ITEM *next = first->next;
		delete first;
		first = next;
	}
}

		
PUBLIC void ACL_LIST::add (ACL_ITEM *item)
{
	item->next = first;
	first = item;
}

/*
	Return true if this uid,gid pair is member of this list.

	It is a member if either the uid or the gid is a member.
*/
PUBLIC bool ACL_LIST::is_member (int uid, int gid) const
{
	ACL_ITEM *it = first;
	while (it != NULL){
		if (it->list != NULL){
			if (it->list->is_member(uid,gid)){
				break;
			}
		}else if (it->type == ITEM_UID && it->id == uid){
			break;
		}else if (it->type == ITEM_GID && it->id == gid){
			break;
		}else if (it->type == ITEM_ALL){
			break;
		}
		it = it->next;
	}
	return it != NULL;
}

/*
	Display the content of a list
*/
PUBLIC void ACL_LIST::dump (const char *indent, FILE *fout)
{
	ACL_ITEM *it = first;
	bool put_comma = false;
	if (it != NULL) fputs (indent,fout);
	while (it != NULL){
		if (put_comma) fputc(',',fout);
		put_comma = true;
		if (it->list != NULL){
			fprintf (fout,"&%s",it->list->name);
		}else if (it->type == ITEM_UID){
			struct passwd *p = getpwuid (it->id);
			if (p == NULL){
				fprintf (fout,"%d",it->id);
			}else{
				fprintf (fout,"%s",p->pw_name);
			}
		}else if (it->type == ITEM_GID){
			struct group *g = getgrgid (it->id);
			if (g == NULL){
				fprintf (fout,"%d",it->id);
			}else{
				fprintf (fout,"%s",g->gr_name);
			}
		}
		it = it->next;
	}
	if (it != first) fputc ('\n',fout);
}

PUBLIC ACL_LISTS::ACL_LISTS()
{
	first = NULL;
}
PUBLIC void ACL_LISTS::add(ACL_LIST *list)
{
	list->next = first;
	first = list;
	
}

PUBLIC ACL_LISTS::~ACL_LISTS()
{
	while (first != NULL){
		ACL_LIST *next = first->next;
		delete first;
		first = next;
	}
}


PUBLIC ACL_CONTEXT::ACL_CONTEXT( ACLTOK &tok, ACL_LISTS &lists)
{
	next = NULL;
	rights = 0;
	if (tok.next() == ACLTOK_ID){
		name = strdup (tok.token);
		if (tok.next() == ACLTOK_OPNBRAC){
			while (tok.next() != ACLTOK_EOF){
				if (tok.type == ACLTOK_CLSBRAC){
					break;
				}else if (tok.cmp("grant")==0){
					grant.parse(tok,lists);
				}else if (tok.cmp("deny")==0){
					deny.parse(tok,lists);
				}else if (tok.cmp("program")==0){
				}else if (tok.cmp("module")==0){
				}else if (tok.cmp("rights")==0){
					bool expectid = true;
					bool opernot = false;	// ! operator seen
					while (tok.next()!=ACLTOK_EOF){
						if (tok.type == ACLTOK_SCOLON) break;
						if (expectid){
							if(tok.type == ACLTOK_NOT){
								opernot = !opernot;
							}else if(tok.type == ACLTOK_ID){
								ACL_RIGHTS r = acl_rights_none;
								if (tok.cmp("overwrite")==0){
									r = may_overwrite;
								}else if (tok.cmp("write")==0){
									r = may_write;
								}else if (tok.cmp("append")==0){
									r = may_append;
								}else if (tok.cmp("trunc")==0){
									r = may_trunc;
								}else if (tok.cmp("read")==0){
									r = may_read;
								}else if (tok.cmp("create")==0){
									r = may_create;
								}else if (tok.cmp("mkdir")==0){
									r = may_mkdir;
								}else if (tok.cmp("unlink")==0){
									r = may_unlink;
								}else if (tok.cmp("rmdir")==0){
									r = may_rmdir;
								}else if (tok.cmp("stat")==0){
									r = may_stat;
								}else if (tok.cmp("readdir")==0){
									r = may_readdir;
								}else if (tok.cmp("rename")==0){
									r = may_rename;
								}else if (tok.cmp("utimes")==0){
									r = may_utimes;
								}else if (tok.cmp("link")==0){
									r = may_link;
								}else if (tok.cmp("symlink")==0){
									r = may_symlink;
								}else if (tok.cmp("chmod")==0){
									r = may_chmod;
								}else if (tok.cmp("readlink")==0){
									r = may_readlink;
								}else if (tok.cmp("ALL")==0){
									r = acl_rights_all;
								}else{
									tok.logerror("Rights definition, invalid privilege %s"
										,tok.token);
								}
								if (opernot){
									rights &= ~r;
								}else{
									rights |= r;
								}
								expectid = false;
							}else{
								tok.logerror("Rights definition, expected a privilege, got %s"
									,tok.token);
							}
						}else if (tok.type != ACLTOK_COMMA){
							tok.logerror("Rights definition, expected a comma, got %s"
									,tok.token);
						}else{
							expectid = true;
						}

					}
				}
			}
		}else{
			tok.logsyntax();
		}
	}else{
		tok.logsyntax();
	}
}

PUBLIC ACL_CONTEXTS::ACL_CONTEXTS()
{
	first = NULL;
}

PUBLIC void ACL_CONTEXTS::add(ACL_CONTEXT *ctx)
{
	// We append at the end instead of inserting, to keep
	// the ordering
	ACL_CONTEXT **pt = &first;
	while (*pt != NULL) pt = &(*pt)->next;
	*pt = ctx;
}

PUBLIC ACL_CONTEXTS::~ACL_CONTEXTS()
{
	while (first != NULL){
		ACL_CONTEXT *next = first->next;
		delete first;
		first = next;
	}
}

PUBLIC ACL_DIRITEM::ACL_DIRITEM (
	ACLTOK &tok,
	ACL_LISTS &lists)
{
	next = NULL;
	if (tok.next() == ACLTOK_ID){
		pattern = strdup (tok.token);
		if (tok.next() == ACLTOK_OPNBRAC){
			while (tok.next() != ACLTOK_EOF){
				if (tok.type == ACLTOK_CLSBRAC){
					break;
				}else if (tok.type == ACLTOK_ID){
					if (tok.cmp ("context")==0){
						contexts.add (new ACL_CONTEXT(tok,lists));
					}else{
						tok.logerror ("expecting context");
					}
				}else{
					tok.logerror ("exporting context");
				}
			}
		}else{
			tok.logsyntax();
		}
	}else{
		tok.logsyntax();
	}
}

PUBLIC ACL_DIRITEM::~ACL_DIRITEM ()
{
	free (pattern);
}


PUBLIC ACL_DIRITEMS::ACL_DIRITEMS()
{
	first = NULL;
}

PUBLIC ACL_DIRITEMS::~ACL_DIRITEMS()
{
	while (first != NULL){
		ACL_DIRITEM *next = first->next;
		delete first;
		first = next;
	}
}

PUBLIC void ACL_DIRITEMS::add (ACL_DIRITEM *d)
{
	// We append at the end instead of inserting, to keep
	// the ordering
	ACL_DIRITEM **pt = &first;
	while (*pt != NULL) pt = &(*pt)->next;
	*pt = d;
}
PUBLIC ACL_DIR::ACL_DIR (
	const char *_path,
	ACL_DIR *_next,
	bool outlog)		// Log error in syslog instead of stderr
{
	aclsyntaxok = true;
	inherit = NULL;
	next = _next;
	path = strdup(_path);
	char tmp[PATH_MAX];
	snprintf (tmp,sizeof(tmp)-1,"%s/..acl",_path);
	FILE *fin = fopen (tmp,"r");
	if (fin != NULL){
		ACLTOK tok (tmp,fin,outlog);
		while (tok.next() != ACLTOK_EOF){
			if (tok.cmp("list")==0){
				ACL_LIST *list = new ACL_LIST (tok,lists);
				lists.add (list);
			}else if (tok.cmp("file")==0){
				ACL_DIRITEM *ditem = new ACL_DIRITEM (tok,lists);
				items.add (ditem);
			}else{
				tok.logerror ("Syntax error");
				break;
			}
		}
		fclose (fin);
		no__acl = false;
		// If there was any error in the ..acl file of a directory
		// then all access is disabled to avoid security breach
		aclsyntaxok = tok.waserror() ? false : true;
	}else{
		no__acl = true;
	}
}

static void aclrights_dumprights( FILE *fout, int rights)
{
	static const char *tbr[]={
		"overwrite",
		"write",
		"append",
		"trunc",
		"read",
		"create",
		"mkdir",
		"unlink",
		"rmdir",
		"stat",
		"readdir",
		"rename",
		"utimes",
		"link",
		"symlink",
		"chmod",
		"readlink",
	};
	bool put_comma = false;
	for (unsigned i=0; i<sizeof(tbr)/sizeof(tbr[0]); i++){
		int r = 1<<i;
		if (rights & r){
			fprintf (fout,"%c%s",put_comma ? ',' : ' ',tbr[i]);
			put_comma = true;
		}
	}
}

PUBLIC void ACL_DIR::dump (FILE *fout)
{
	ACL_DIRITEM *it = items.first;
	while (it != NULL){
		fprintf (fout,"file %s{\n",it->pattern);
		ACL_CONTEXT *ctx = it->contexts.first;
		while (ctx != NULL){
			fprintf (fout,"\tcontext %s{\n",ctx->name);
			fprintf (fout,"\t\tgrant{\n");
			ctx->grant.dump ("\t\t\t",stdout);
			fprintf (fout,"\t\t}\n");
			fprintf (fout,"\t\tdeny{\n");
			ctx->deny.dump ("\t\t\t",stdout);
			fprintf (fout,"\t\t}\n");
			fprintf (fout,"\t\trights");
			aclrights_dumprights (fout,ctx->rights);
			fprintf (fout,";\n");
			ctx = ctx->next;
			fprintf (fout,"\t}\n");
		}
		fprintf (fout,"}\n");
		it = it->next;
	}
}

/*
	Compute the access rights for a uid,gid,pid,file combination
	Return the bit set ACL_RIGHTS.
*/
PUBLIC int ACL_DIR::getrights(
	int uid,
	int gid,
	int pid,
	const char *file)
{
	int rights = 0;
	ACL_DIRITEM *ditem = items.first;
	//n_debug ("getrights dir %s\n",path);
	while (ditem != NULL){
		//n_debug ("matching :%s: :%s:\n",ditem->pattern,file);
		if (strcmp(ditem->pattern,file)==0
			|| strcmp(ditem->pattern,"*")==0){
			ACL_CONTEXT *ctx = ditem->contexts.first;
			while (ctx != NULL){
				if (ctx->deny.is_member(uid,gid)){
					rights &= ~ctx->rights;
				}else if (ctx->grant.is_member(uid,gid)){
					rights |= ctx->rights;
				}
				ctx = ctx->next;
			}
			break;
		}
		ditem = ditem->next;
	}
	//n_debug ("getrights end\n");
	return rights;
}

static ACL_DIR *first;

ACL_DIR *aclrights_locatedir (
	const char *dir,
	bool outlog)
{
	ACL_DIR *d = first;
	while (d != NULL){
		if (strcmp(d->path,dir)==0){
			break;
		}
		d = d->next;
	}
	if (d == NULL){
		d = new ACL_DIR (dir,first,outlog);
		first = d;
		if (d->no__acl){
			// Locate the parent
			char tmp[strlen(dir)+1];
			strcpy (tmp,dir);
			char *pt=strrchr(tmp,'/');
			if (pt!=NULL){
				*pt = '\0';
				ACL_DIR *parent = aclrights_locatedir (tmp,outlog);
				if (parent->inherit != NULL){
					d->inherit = parent->inherit;
				}else{
					d->inherit = parent;
				}
			}
		}
	}
	return d;
}


/*
	Get the access rights for a uid,gid,pid,dir,file combination
*/
void aclrights_get (
	int uid,
	int gid,
	int pid,
	const char *dir,
	const char *file,
	int &rights)
{
	rights = 0;
	// ..acl is protected. Can't be seen, read, stat, whatever
	if (strcmp(file,"..acl")!=0){
		ACL_DIR *d = aclrights_locatedir (dir,true);
		if (d == NULL || !d->aclsyntaxok){
			// All access disabled
			rights = 0;
		}else{
			rights = d->getrights (uid,gid,pid,file);
		}
	}
}

/*
	Reload the acl configuration for a directory
	(forget the cached one in fact)
*/
void aclrights_reload (const char *dir)
{
	ACL_DIR **pt = &first;
	while (*pt != NULL){
		ACL_DIR *d = *pt;
		if (strcmp(d->path,dir)==0){
			*pt = d->next;
			delete d;
			break;
		}
		pt = &d->next;
	}
	// We must walk the directory inheriting from this one
}


/*
	Report the access rights for a given user/group
*/
void aclrights_testacl (
	const char *dir,
	const char *file,
	const char *usr,
	const char *grp)
{
	ACL_DIR *d = aclrights_locatedir (dir,false);
	if (d == NULL){
		fprintf (stderr,"No acls for directory %s\n",dir);
	}else{
		printf ("Acl for directory %s: ",dir);
		if (d->inherit != NULL){
			printf ("Inherited from directory %s",d->inherit->path);
			d = d->inherit;
		}
		printf ("\n");
		struct passwd *p = getpwnam (usr);
		struct group * g = getgrnam (grp);
		if (p == NULL){
			fprintf (stderr,"Non existent user %s\n",usr);
		}else if (g == NULL){
			fprintf (stderr,"Non existent group %s\n",grp);
		}else{
			d->dump (stdout);
			int rights = d->getrights (p->pw_uid,g->gr_gid,-1,file);
			printf ("Rights: ");
			aclrights_dumprights (stdout,rights);
			printf ("\n");
		}
	}
}

/*
	Check if an ..acl file is ok in a directory.
	Return -1 if any error
*/
int aclrights_checkacl(const char *dir)
{
	int ret = -1;
	ACL_DIR *d = aclrights_locatedir (dir,false);
	if (d == NULL){
		fprintf (stderr,"No acls for directory %s\n",dir);
	}else{
		if (d->inherit != NULL){
			d = d->inherit;
		}
		if (d->aclsyntaxok){
			ret = 0;
		}else{
			fprintf (stderr,"Acls for directory %s are invalid\n",dir);
		}
	}
	return ret;
}

