#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "litefs.h"
#include "litefs.p"

static void usage()
{
	fprintf (stderr,"litefs-proxy version 1.0\n"
		"\n"
		"litefs-proxy --socket sock_path [ --fork ] [ --debug] [ --exec command ...]\n"
		"litefs-proxy --pipe [ --fork] [ --debug] [ --exec command ...]\n");
}

static bool debug=false;

static void n_debug (const char *ctl, ...)
{
	if (debug){
		va_list list;
		va_start (list,ctl);
		vfprintf (stderr,ctl,list);
		va_end (list);
	}
}


static int proxy_doreq (
	int client_in,
	int client_out,
	int proxy_in,
	int proxy_out)
{
	int ret = -1;
	LITE_REQUEST_DATA data;
	int n = read (client_in,&data,sizeof(data));
	if (n >= (int)sizeof(LITE_REQUEST)){
		int len = n - sizeof(LITE_REQUEST);
		while (len < data.len){
			int chunk = read (client_in,(char*)data.buf+len
				,sizeof(data.buf)-len);
			n_debug ("request read %d len %d chunk %d\n",data.cmd,len,chunk);
			if (chunk > 0){
				len += chunk;
			}else{
				break;
			}
		}
		n = sizeof(LITE_REQUEST)+len;
		n_debug ("-proxy write n=%d cmd=%d\n",n,data.cmd);
		if (write (proxy_out,&data,n) == n){
			LITE_REPLY_DATA rep;
			n = read (proxy_in,&rep,sizeof(rep));
			n_debug ("-proxy read %d len %d\n",n,rep.len);
			if (n >= (int)sizeof(LITE_REPLY)){
				len = n - sizeof(LITE_REPLY);
				while (len < rep.len){
					int chunk = read (proxy_in,(char*)rep.buf+len
						,sizeof(rep.buf)-len);
					n_debug ("reply read len %d chunk %d\n",len,chunk);
					if (chunk > 0){
						len += chunk;
					}else{
						break;
					}
				}
				n = len + sizeof(LITE_REPLY);
				if (write (client_out,&rep,n)==n){
					ret = 0;
				}
			}
		}
	}
	return ret;
}



static int proxy_fdset (fd_set &fdin, int maxfd, int handle)
{
	if (handle != -1){
		FD_SET (handle,&fdin);
		if (handle > maxfd) maxfd = handle;
	}
	return maxfd;
}

/*
	Process message from the back-end.
	Return -1 if any error.

	At some point, we will introduce some optimisation to
	enhance the performance of litefs over high latency link.
	These enhancements include
		performing LITE_WRITE without waiting for an immediate
		answer (but report success to the caller). This
		will allow streaming of write request

		performs read ahread for LITE_READ.

	Those tricks will introduce asyncrhronous answers from
	the back-end and will be managed here.
*/
static int proxy_processopt (int proxyin, int proxyout)
{
	// For now, no message expected
	return -1;
}

static void proxy_loop (
	int sock,
	int pipein,
	int pipeout,
	int proxyin,
	int proxyout)
{
	int tbfd[1024];
	int nbfd = 0;
	fd_set fdin;
	FD_ZERO (&fdin);
	int maxfd = 0;
	maxfd = proxy_fdset (fdin,maxfd,sock);
	maxfd = proxy_fdset (fdin,maxfd,pipein);
	maxfd = proxy_fdset (fdin,maxfd,pipeout);
	maxfd = proxy_fdset (fdin,maxfd,proxyin);
	while (1){
		fd_set in;
		memcpy (&in,&fdin,sizeof(in));
		n_debug ("select maxfd %d\n",maxfd);
		if (select (maxfd+1,&in,NULL,NULL,NULL)!=-1){
			int orig_nbfd = nbfd;
			if (sock != -1 && FD_ISSET (sock,&in)){
				struct sockaddr_un unc;
				size_t len = sizeof(unc);
				unc.sun_family = AF_UNIX;
				int s = accept (sock,(struct sockaddr*)&unc,&len);
				tbfd[nbfd++] = s;
				FD_SET (s,&fdin);
				if (s >= maxfd) maxfd = s + 1;
			}
			if (pipein != -1 && FD_ISSET(pipein,&in)){
				if (proxy_doreq (pipein,pipeout,proxyin,proxyout) == -1){
					break;
				}
			}
			if (FD_ISSET(proxyin,&in)){
				if (proxy_processopt(proxyin,proxyout)==-1){
					break;
				}
			}
			int *ptfd = tbfd;
			bool redo_fdin = false;
			for (int i=0; i< orig_nbfd; i++, ptfd++){
				int fd = *ptfd;
				if (FD_ISSET(fd,&in)){
					if (proxy_doreq (fd,fd,proxyin,proxyout) == -1){
						close (fd);
						redo_fdin = true;
						*ptfd = -1;
					}
				}
			}
			if (redo_fdin){
				FD_ZERO (&fdin);
				maxfd = 0;
				maxfd = proxy_fdset (fdin,maxfd,sock);
				maxfd = proxy_fdset (fdin,maxfd,pipein);
				maxfd = proxy_fdset (fdin,maxfd,proxyin);
				ptfd = tbfd;
				int *copyfd = tbfd;
				for (int i=0; i<nbfd; i++, ptfd++){
					int fd = *ptfd;
					if (fd != -1){
						maxfd = proxy_fdset (fdin,maxfd,fd);
						*copyfd++ = fd;
					}
				}
				nbfd = (int)(copyfd - tbfd);
			}
		}
	}
}

/*
	Wait for a confirmation status.
	Return -1 if the original request failed.
*/
static int litefs_receive(int fd)
{
	int ret = -1;
	LITE_REPLY rep;
	int len = read (fd,&rep,sizeof(rep));
	if (len == sizeof(LITE_REPLY)){
		if (rep.status == LITE_STATUS_OK || rep.status == LITE_STATUS_SEEK){
			ret = rep.data;
			errno = 0;
		}else{
			errno = rep.errcode;
		}
	}
	return ret;
}


static void litefs_sendreq (int fd, LITE_REQUEST &req, int len)
{
	req.len = len;
	int ret = write (fd,&req,sizeof(LITE_REQUEST)+len);
	if (ret==-1){
		perror ("sendmsg");
	}
}

int main (int argc, char *argv[])
{
	int ret = -1;
	if (argc == 1){
		usage();
	}else{
		int sock = -1;
		int pipein = -1;
		int pipeout = -1;
		int proxyin = -1;
		int proxyout = -1;
		bool dofork = false;
		for (int i=1; i<argc; i++){
			if (strcmp(argv[i],"--socket")==0){
				i++;
				sock = litefsd_server(argv[i]);
			}else if (strcmp(argv[i],"--debug")==0){
				debug = true;
			}else if (strcmp(argv[i],"--fork")==0){
				setsid();	// Make sure that we are not killed when
							// the session leader die
				signal (SIGHUP,SIG_IGN);
				dofork = true;
			}else if (strcmp(argv[i],"--pipe")==0){
				pipein = 0;
				pipeout = 1;
			}else if (strcmp(argv[i],"--exec")==0){
				if (i == argc -1){
					usage();
					exit (-1);
				}else{
					int tbin[2],tbout[2];
					if (pipe(tbin)==-1 || pipe(tbout)==-1){
						fprintf (stderr,"Can't setup pipe (%s)\n",strerror(errno));
					}else{
						int pid = fork();
						if (pid == -1){
							fprintf (stderr,"Can't fork new process (%s)\n"
								,strerror(errno));
						}else if (pid == 0){
							close (tbout[1]);
							close (tbin[0]);
							dup2 (tbin[1],1);
							dup2 (tbout[0],0);
							close (tbout[0]);
							close (tbin[1]);
							char line[2000];
							line[0] = '\0';
							for (int j=i+1; j<argc; j++){
								strcat (line,argv[j]);
								strcat (line," ");
							}
							system (line);
							_exit (0);
						}else{
							close (tbout[0]);
							close (tbin[1]);
							proxyin = tbin[0];
							proxyout = tbout[1];
						}
						i=argc-1;
					}
				}
			}else{
				usage();
				exit (-1);
			}
		}
		if ((sock != -1 || pipein != -1) && proxyin != -1){
			if (dofork){
				// First, make sure the link is established
				// before forking. This allows ssh to request
				// a password
				LITE_REQUEST req;
				req.cmd = LITE_VERSION;
				req.mode = LITEFS_PROTOCOL_VERSION;
				litefs_sendreq (proxyout,req,0);
				ret = litefs_receive(proxyin);
				if (ret == -1){
					fprintf (stderr,"Incompatible litefs backend\n");
					return -1;
				}
				if (fork() != 0) exit(0);
			}
			proxy_loop(sock,pipein,pipeout,proxyin,proxyout);
		}else{
			usage();
		}
	}
	return ret;
}




