/*
	This is a LITE client.

	It maps various filesystems into the client home/n directory.
	It deals with the litefsld (local) server to gain access to files.
	The later is hooked to litefsrd (remote) server which implements
	the request.

	litefsld (local daemon) stands as a proxy to the remote litefsrd.
	This avoids opening a new remote connection for each new client.

	Note that the litefs module might be used to talk to potentially
	many different litefsld. Here are some example

	ftpfsd: Virtualise an ftp session as a file system. Unlike other
		ftp vfs solution, this one would be usable even in shell
		scripts...

	The implementation is pretty much the same as aclfs, except that
	aclfs banks on the ability to pass file handle though the protocol
	(unix domain socket). At some point, the two module might be
	merged.

	The intent usage of litefs is this
	
	You start this command

		litefsld --socket ~/.ahost_socket ssh -l user a_host litefsrd

	Then you can browse the remote site using file system tools
	even if you do not have normal file system access (smb, nfs, ...).
	On each side of the connection, all is handled in userland and
	requires no privilege.
*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <wait.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>
#include <limits.h>
#include <unistd.h>
#include <sys/time.h>
#include <utime.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "litefs.h"
#include "remaudio.h"
#include "dispatch.h"

struct LITEFS_CONF {
	const char *volume;
	const char *sockpath;
	const char *startcmd;
};

static int nbdef=0;
static LITEFS_CONF *conf;

static const char *litefs_copyword (char *dst, const char *src)
{
	while (isspace (*src)) src++;
	while (*src > ' ') *dst++ = *src++;
	*dst = '\0';
	return src;
}

static void litefs_readconf ()
{
	if (conf == NULL){
		char path[PATH_MAX];
		snprintf (path,sizeof(path)-1,"%s/.litefs",_dispatch_gethome());
		FILE *fin = fopen (path,"r");
		if (fin != NULL){
			char buf[1000];
			conf = (LITEFS_CONF*)malloc (100*sizeof(LITEFS_CONF));
			while (fgets(buf,sizeof(buf)-1,fin)!=NULL){
				char volume[PATH_MAX],sockpath[PATH_MAX];
				const char *pt = litefs_copyword (volume,buf);
				pt = litefs_copyword (sockpath,pt);
				while (isspace(*pt)) pt++;
				if (volume[0] != '\0' && sockpath[0] != '\0'){
					conf[nbdef].volume = strdup(volume);
					char fullpath[PATH_MAX];
					if (sockpath[0] != '/'){
						snprintf (fullpath,sizeof(fullpath)-1,"%s/%s"
							,_dispatch_gethome(),sockpath);
					}else{
						strcpy (fullpath,sockpath);
					}
					conf[nbdef].sockpath = strdup(fullpath);
					conf[nbdef].startcmd = strdup(pt);
					nbdef++;
				}
			}
			fclose (fin);
		}
	}
}

struct LITEFS_INFO{
	char volume[PATH_MAX];		// Sub-directory in $HOME/n
	char relpath[PATH_MAX];		// relative path in $HOME/n/volume/
	const char *sockpath;		// Socket to use to reach litefsld
	const char *startcmd;		// How to start the server if not already
								// running
	char virtpath[PATH_MAX];	// absolute virtual path
};

/*
	Check if a volume is configured for the lite file system
*/
static int litefs_chkconf (LITEFS_INFO &info)
{
	litefs_readconf();
	int ret = -1;
	for (int i=0; i<nbdef; i++){
		if (strcmp(info.volume,conf[i].volume)==0){
			info.sockpath = conf[i].sockpath;
			info.startcmd = conf[i].startcmd;
			ret = 0;
			break;
		}
	}
	return ret;
}

static int lite_fd=-1;	// Socket to do request

static short lookupfd[1024];

/*
	Wait to receive a file handle
	We receive a handle used to do further request to the litefsld server.

	For now, to avoid confusing the application, we translate this file handle
	by allocating a real one from the OS.

	Return the file handle or -1.
*/
static int litefs_receivefd()
{
	int ret = -1;
	LITE_REPLY rep;
	int len = read (lite_fd,&rep,sizeof(rep));
	if (len == sizeof(LITE_REPLY)){
		if (rep.status == LITE_STATUS_HANDLE){
			int fd = open ("/dev/null",O_RDWR);
			// fprintf (stderr,"open /dev/null %d\n",fd);
			lookupfd[fd] = rep.data;
			ret = fd;
		}else{
			errno = rep.errcode;
		}
	}
	return ret;
}

/*
	Wait for a confirmation status.
	Return -1 if the original request failed.
*/
static int litefs_receive()
{
	int ret = -1;
	LITE_REPLY rep;
	int len = read (lite_fd,&rep,sizeof(rep));
	if (len == sizeof(LITE_REPLY)){
		if (rep.status == LITE_STATUS_OK || rep.status == LITE_STATUS_SEEK){
			ret = rep.data;
			errno = 0;
		}else{
			errno = rep.errcode;
		}
	}
	return ret;
}

/*
	Wait for a some data
	Return -1 if the original request failed.
*/
static int litefs_receive_data(void *buf, int len, LITE_STATUS expect)
{
	int ret = -1;
	LITE_REPLY_DATA rep;
	int lenread = read (lite_fd,&rep,sizeof(rep));
	// fprintf (stderr,"lenread %d  %u\n",lenread,sizeof(LITE_REPLY));
	if (lenread >= (int)sizeof(LITE_REPLY)){
		if (rep.status == expect){
			lenread -= sizeof(LITE_REPLY);
			if (lenread > 0) memcpy (buf,rep.buf,lenread);
			while (rep.len > lenread){
				int chunk = read (lite_fd,(char*)buf+lenread,len-lenread);
				if (chunk >= 0){
					lenread += chunk;
				}else{
					break;
				}
			}
			ret = lenread;
			errno = 0;
		}else{
			errno = rep.errcode;
		}
	}
	return ret;
}

/*
	Wait for a some data
	Return -1 if the original request failed.
*/
static int litefs_receive_data(void *buf, int len)
{
	return litefs_receive_data (buf,len,LITE_STATUS_OK);
}

/*
	Wait for a confirmation of a stat request.
	Return -1 if the original request failed.
*/
static int litefs_receive_stat (struct stat &buf)
{
	int ret = -1;
	int len = litefs_receive_data (&buf,sizeof(struct stat),LITE_STATUS_STAT);
	// fprintf (stderr,"receive_stat %d %u\n",len,sizeof(struct stat));
	if (len == sizeof(struct stat)){
		ret = 0;
	}
	return ret;
}

static void litefs_sendreq (LITE_REQUEST &req, int len)
{
	req.len = len;
	int ret = __libc_write (lite_fd,&req,sizeof(LITE_REQUEST)+len);
	if (ret==-1){
		perror ("sendmsg");
	}
}

static int litefs_connect (const char *sockpath, bool showerror)
{
	int ret = -1;
	if (lite_fd != -1){
		ret = 0;
	}else{
		int fd =  socket (AF_UNIX,SOCK_STREAM,0);
		if (fd == -1){
			if (showerror) perror("socket client");
		}else{
			struct sockaddr_un un;
			un.sun_family = AF_UNIX;
			strcpy (un.sun_path,sockpath);
			int s = connect(fd,(struct sockaddr*)&un,sizeof(un));
			if (s == -1){
				if (showerror) perror("connect");
			}else{
				lite_fd = fd;
				LITE_REQUEST req;
				req.cmd = LITE_VERSION;
				req.mode = LITEFS_PROTOCOL_VERSION;
				litefs_sendreq (req,0);
				ret = litefs_receive();
				if (ret != 0){
					close (fd);
					lite_fd = -1;
				}
			}
		}
	}
	return ret;
}

/*
	Simplify a path by stripping .. and .
	Return -1 if path is invalid, 0 if ok
	No validation are made on disk (for directory existence).

	If result is empty, replace by "." or "/" if the path was  absolute.
*/
int litefs_simplifypath (const char *path, char *newp)
{
	int ok = 0;
	char *debnewp = newp;
	char *lastvalid[20];	/* Last member other than . and .. */
	int nbvalid = 0;
	const bool isabs = path[0] == '/';
	char tmp[PATH_MAX];
	if (path == newp){
		strcpy (tmp,path);
		path = tmp;
	}
	while (path[0] == '.' && path[1] == '/') path += 2;
	if (path[0] != '/' && path[0] != '.'){
		lastvalid[nbvalid++] = newp;
	}
	while (ok == 0 && *path	!= '\0'){
		if (path[0] == '/'
			&& path[1] == '.'
			&& (path[2] == '/' || path[2] == '\0')){
			path += 2;
		}else if (path[0] == '/'
			&& path[1] == '.'
			&& path[2] == '.'
			&& (path[3] == '/' || path[3] == '\0')){
			if (nbvalid > 0){
				newp = lastvalid[--nbvalid];
				/* x/../y -> y and not /y */
				if (newp == debnewp
					&& !isabs
					&& path[3] == '/') path++;
			}else if (isabs){
				ok = -1;
				break;
			}else{
				*newp++ = '/';
				*newp++ = '.';
				*newp++ = '.';
				*newp = '\0';
			}
			path += 3;
		}else{
			if (*path == '/') lastvalid[nbvalid++] = newp;
			*newp++ = *path++;
		}
	}
	*newp = '\0';
	if (debnewp[0] == '\0'){
		debnewp[0] = isabs ? '/' : '.';
		debnewp[1] = '\0';
	}
	return ok;
}

/*
	Translate a logical path if it is part of the personnal mount area
	($HOME/n)
*/
static int litefs_translate (
	const char *path,
	LITEFS_INFO &info,
	LITE_REQUEST_DATA &req,
	int &len_request)
{
	len_request = 0;
	int lenn;
	const char *pathn = _dispatch_getpathn(lenn);
	if (pathn == NULL) return -1;
	if (path[0] != '/'){
		const char *curpath = _dispatch_getcwd();
		snprintf (info.virtpath,sizeof(info.virtpath)-1,"%s/%s",curpath,path);
		path = info.virtpath;
	}else{
		strcpy (info.virtpath,path);
	}
	if (strstr(path,"..")!=NULL){
		litefs_simplifypath (info.virtpath,info.virtpath);
		path = info.virtpath;
	}
	int ret = -1;

	// fprintf (stderr,"pathn = :%s: <> :%s:\n",pathn,path);
	if (strncmp(pathn,path,lenn)==0){
		path += lenn;
		char *volume = info.volume;
		while (*path != '\0' && *path != '/'){
			*volume++ = *path++;
		}
		*volume = '\0';
		strcpy (info.relpath,path);
		// fprintf (stderr,"volume :%s: :%s:\n",info.volume,info.relpath);
		if (litefs_chkconf (info)!=-1){
			int ok = litefs_connect(info.sockpath,false);
			if (ok == -1){
				_dispatch_exec (info.startcmd);
				ok = litefs_connect(info.sockpath,true);
			}
			if (ok != -1){
				strcpy (req.buf,info.relpath);
				len_request = strlen(info.relpath)+1;
				req.mode = 0;
				req.perm = 0;
				ret = 0;
			}
		}
	}
	return ret;
}

static int litefs_open (const char *path, int mode, int perm)
{
	int ret = DISPATCH_DONTCARE;
	LITEFS_INFO info;
	LITE_REQUEST_DATA req;
	int len;
	if (litefs_translate (path,info,req,len)!=-1){
		// fprintf (stderr,"open :%s: %d %d\n",info.relpath,mode,perm);
		req.cmd = LITE_OPEN;
		req.mode = mode;
		req.perm = perm;
		litefs_sendreq (req,len);
		ret = litefs_receivefd();
	}
	return ret;
}

static int litefs_close (int fd)
{
	LITE_REQUEST req;
	req.cmd = LITE_CLOSE;
	req.handle = lookupfd[fd];
	req.mode = 0;
	req.perm = 0;
	litefs_sendreq (req,0);
	int ret = litefs_receive();
	__libc_close (fd);
	return ret;
}

static int litefs_read (int fd, void *buf, unsigned size)
{
	int ret = 0;
	LITE_REQUEST req;
	req.cmd = LITE_READ;
	req.handle = lookupfd[fd];
	req.perm = 0;
	int pos = 0;
	int received = 0;
	// fprintf (stderr,"litefs_read %d %d\n",req.handle,size);
	while (size > 0){
		int chunk = 32000;
		if (size <= 32000) chunk = size;
		size -= chunk;
		req.mode = chunk;
		litefs_sendreq (req,0);
		ret = litefs_receive_data((char*)buf+pos,chunk);
		// fprintf (stderr."litefs_read ret=%d\n",ret);
		if (ret <= 0) break;
		received += ret;
		pos += chunk;
	}
	// fprintf (stderr,"litefs_read size=%d received=%d errno=%d\n",size,received,errno);
	return received > 0 ? received : ret;
}

static int litefs_write (int fd, const void *buf, unsigned size)
{
	int ret = 0;
	LITE_REQUEST_DATA req;
	req.cmd = LITE_WRITE;
	req.handle = lookupfd[fd];
	req.perm = 0;
	int pos = 0;
	int sent = 0;
	while (size > 0){
		// fprintf (stderr,"write size=%d pos=%d\n",size,pos);
		int chunk = 1500;
		if (size <= 1500) chunk = size;
		size -= chunk;
		req.len = chunk;
		memcpy (req.buf,(char*)buf+pos,chunk);
		litefs_sendreq (req,chunk);
		ret = litefs_receive ();
		if (ret <= 0) break;
		sent += chunk;
		pos += chunk;
	}
	return sent > 0 ? sent : ret;
}

static off_t litefs_lseek (int fd, off_t offset, int whence)
{
	LITE_REQUEST req;
	req.cmd = LITE_SEEK;
	req.handle = lookupfd[fd];
	req.mode = offset;
	req.perm = whence;
	litefs_sendreq (req,0);
	return litefs_receive();
}
	


static int litefs_mkdiranc (const char *path)
{
	int ret = __mkdir (path,0755);
	if (ret == -1){
		char tmp[strlen(path)+1];
		strcpy (tmp,path);
		char *pt = strrchr(tmp,'/');
		if (pt != NULL){
			*pt = '\0';
			if (litefs_mkdiranc (tmp) != -1) ret = __mkdir (path,0755);
		}
	}
	return ret;
}


static int litefs_chdir (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	LITEFS_INFO info;
	LITE_REQUEST_DATA req;
	int len;
	if (litefs_translate (path,info,req,len)!=-1){
		req.cmd = LITE_CHDIR;
		litefs_sendreq (req,len);
		ret = litefs_receive();
		// fprintf (stderr,"chdir :%s: :%s: -> %d\n",path,info.virtpath,ret);
		if (ret != -1){
			// We have to set the new current dir in $HOME/n
			if (__chdir (info.virtpath) == -1){
				litefs_mkdiranc (info.virtpath);
				__chdir (info.virtpath);
			}
		}
	}
	return ret;
}

static struct DIR * litefs_opendir (const char *path)
{
	struct DIR *ret = NULL;
	LITEFS_INFO info;
	LITE_REQUEST_DATA req;
	int len;
	if (litefs_translate (path,info,req,len)!=-1){
		// fprintf (stderr,"litefs_opendir :%s:\n",path);
		req.cmd = LITE_OPENDIR;
		litefs_sendreq (req,len);
		int fd = litefs_receivefd();
		if (fd != -1){
			ret = (DIR*)fd;
		}
	}
	return ret;
}

static struct dirent *litefs_readdir (struct DIR *dir)
{
	// fprintf (stderr,"litefs_readdir\n");
	int fd = (int)dir;
	fd = lookupfd[fd];
	LITE_REQUEST req;
	req.cmd = LITE_READDIR;
	req.handle = fd;
	req.mode = 0;
	req.perm = 0;
	litefs_sendreq (req,0);
	static struct dirent ent;
	int len = litefs_receive_data(&ent,sizeof(ent));
	struct dirent *ret = &ent;
	if (len != (int)sizeof(ent)){
		ret = NULL;
	}else{
		//fprintf (stderr,"readdir :%s: ino %ld reclen %d\n",ent.d_name,ent.d_ino
		//	,ent.d_reclen);
	}
	return ret;
}


static int litefs_closedir (struct DIR *dir)
{
	int fd = (int)dir;
	__libc_close (fd);
	fd = lookupfd[fd];
	LITE_REQUEST req;
	req.cmd = LITE_CLOSEDIR;
	req.handle = fd;
	req.mode = 0;
	req.perm = 0;
	litefs_sendreq (req,0);
	return litefs_receive();
}
static int litefs_rename (const char *old, const char *newp)
{
	int ret = DISPATCH_DONTCARE;
	LITEFS_INFO oldinfo, newinfo;
	LITE_REQUEST_DATA old_req,new_req;
	int old_len, new_len;
	
	int oldf = litefs_translate (old,oldinfo,old_req,old_len);
	int newf = litefs_translate (newp,newinfo,new_req,new_len);
	// fprintf (stderr,"litefs_rename %d %d :%s: :%s:\n",oldf,newf,oldinfo.volume,newinfo.volume);
	if (oldf != -1 && newf !=-1
		&& strcmp(oldinfo.volume,newinfo.volume)==0){
		old_req.cmd = LITE_RENAME;
		strcpy (old_req.buf+old_len,newinfo.relpath);
		litefs_sendreq (old_req,old_len+new_len);
		ret = litefs_receive();
	}else if (oldf != -1 || newf != -1){
		errno = EXDEV;
		ret = -1;
	}
	return ret;
}
static int litefs_link (const char *old, const char *newp)
{
	int ret = DISPATCH_DONTCARE;
	LITEFS_INFO oldinfo, newinfo;
	LITE_REQUEST_DATA old_req,new_req;
	int old_len, new_len;
	
	int oldf = litefs_translate (old,oldinfo,old_req,old_len);
	int newf = litefs_translate (newp,newinfo,new_req,new_len);
	if (oldf != -1 && newf !=-1
		&& strcmp(oldinfo.volume,newinfo.volume)==0){
		old_req.cmd = LITE_LINK;
		strcpy (old_req.buf+old_len,newinfo.relpath);
		litefs_sendreq (old_req,old_len+new_len);
		ret = litefs_receive();
	}else if (oldf != -1 || newf != -1){
		errno = EXDEV;
		ret = -1;
	}
	return ret;
}

static int litefs_symlink (const char *old, const char *newp)
{
	int ret = DISPATCH_DONTCARE;
	LITEFS_INFO newinfo;
	LITE_REQUEST_DATA new_req;
	int new_len;
	
	if (litefs_translate (newp,newinfo,new_req,new_len) !=-1){
		new_req.cmd = LITE_SYMLINK;
		strcpy (new_req.buf+new_len,old);
		int old_len = strlen (old);
		litefs_sendreq (new_req,old_len+new_len+1);
		ret = litefs_receive();
	}
	return ret;
}

static int litefs_unlink (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	LITEFS_INFO info;
	LITE_REQUEST_DATA req;
	int len;
	if (litefs_translate (path,info,req,len)!=-1){
		// fprintf (stderr,"unlink :%s:\n",info.realpath);
		req.cmd = LITE_UNLINK;
		litefs_sendreq (req,len);
		ret = litefs_receive();
	}
	return ret;
}
static int litefs_rmdir (const char *path)
{
	int ret = DISPATCH_DONTCARE;
	LITEFS_INFO info;
	LITE_REQUEST_DATA req;
	int len;
	if (litefs_translate (path,info,req,len)!=-1){
		req.cmd = LITE_RMDIR;
		litefs_sendreq (req,len);
		ret = litefs_receive();
	}
	return ret;
}
static int litefs_mkdir(const char *path, int mode)
{
	int ret = DISPATCH_DONTCARE;
	LITEFS_INFO info;
	LITE_REQUEST_DATA req;
	int len;
	if (litefs_translate (path,info,req,len)!=-1){
		req.cmd = LITE_MKDIR;
		req.mode = mode;
		litefs_sendreq (req,len);
		ret = litefs_receive();
	}
	return ret;
}

static int litefs_stat (int ver, const char *fname, struct stat *buf)
{
	int ret = DISPATCH_DONTCARE;
	LITEFS_INFO info;
	LITE_REQUEST_DATA req;
	int len;
	if (litefs_translate (fname,info,req,len)!=-1){
		// fprintf (stderr,"stat fname :%s: relpath :%s:\n",fname,info.relpath);
		req.cmd = LITE_STAT;
		litefs_sendreq (req,len);
		ret = litefs_receive_stat(*buf);
		// We make sure the volumes appears as being on another device
		// so backups (tar) won't follow the mount
		if (ret == 0) buf->st_dev = makedev(major(buf->st_dev)+100,minor(buf->st_dev));
		// fprintf (stderr,"Receive stat %d %o\n",ret,buf->st_mode);
	}
	return ret;
}
static int litefs_fstat (int ver, int fd, struct stat *buf)
{
	LITE_REQUEST req;
	req.cmd = LITE_FSTAT;
	req.handle = lookupfd[fd];
	req.perm = 0;
	req.mode = 0;
	litefs_sendreq (req,0);
	int ret = litefs_receive_stat(*buf);
	// We make sure the volumes appears as being on another device
	// so backups (tar) won't follow the mount
	if (ret == 0) buf->st_dev = makedev(major(buf->st_dev)+100,minor(buf->st_dev));
	//fprintf (stderr,"litefs_fstat %d -> %d %d\n",fd,req.handle,ret);
	return ret;
}
static int litefs_lstat (int ver, const char *fname, struct stat *buf)
{
	int ret = DISPATCH_DONTCARE;
	LITEFS_INFO info;
	LITE_REQUEST_DATA req;
	int len;
	if (litefs_translate (fname,info,req,len)!=-1){
		// fprintf (stderr,"lstat fname :%s: relpath :%s:\n",fname,info.relpath);
		req.cmd = LITE_LSTAT;
		litefs_sendreq (req,len);
		ret = litefs_receive_stat(*buf);
		// We make sure the volumes appears as being on another device
		// so backups (tar) won't follow the mount
		if (ret == 0) buf->st_dev = makedev(major(buf->st_dev)+100,minor(buf->st_dev));
		// fprintf (stderr,"Receive lstat %d %o\n",ret,buf->st_mode);
	}
	return ret;
}

static int litefs_chmod (const char *fname, mode_t mode)
{
	int ret = DISPATCH_DONTCARE;
	LITEFS_INFO info;
	LITE_REQUEST_DATA req;
	int len;
	if (litefs_translate (fname,info,req,len)!=-1){
		req.cmd = LITE_CHMOD;
		req.mode = mode;
		litefs_sendreq (req,len);
		ret = litefs_receive();
	}
	return ret;
}

static int litefs_fchmod (int fd, mode_t mode)
{
	LITE_REQUEST req;
	req.cmd = LITE_FCHMOD;
	req.handle = lookupfd[fd];
	req.mode = mode;
	req.perm = 0;
	litefs_sendreq (req,0);
	return litefs_receive();
}
static int litefs_fchown (int fd, uid_t uid, gid_t gid)
{
	LITE_REQUEST req;
	req.cmd = LITE_FCHOWN;
	req.handle = lookupfd[fd];
	req.mode = uid;
	req.perm = gid;
	litefs_sendreq (req,0);
	return litefs_receive();
}

static int litefs_chown (const char *fname, uid_t uid, gid_t gid)
{
	int ret = DISPATCH_DONTCARE;
	LITEFS_INFO info;
	LITE_REQUEST_DATA req;
	int len;
	if (litefs_translate (fname,info,req,len)!=-1){
		req.cmd = LITE_CHOWN;
		req.mode = uid;
		req.perm = gid;
		litefs_sendreq (req,len);
		ret = litefs_receive();
	}
	return ret;
}
static int litefs_lchown (const char *fname, uid_t uid, gid_t gid)
{
	int ret = DISPATCH_DONTCARE;
	LITEFS_INFO info;
	LITE_REQUEST_DATA req;
	int len;
	if (litefs_translate (fname,info,req,len)!=-1){
		req.cmd = LITE_LCHOWN;
		req.mode = uid;
		req.perm = gid;
		litefs_sendreq (req,len);
		ret = litefs_receive();
	}
	return ret;
}


static int litefs_readlink (const char *fname, char *buf, size_t lenbuf)
{
	int ret = DISPATCH_DONTCARE;
	LITEFS_INFO info;
	LITE_REQUEST_DATA req;
	int len;
	if (litefs_translate (fname,info,req,len)!=-1){
		req.cmd = LITE_READLINK;
		litefs_sendreq (req,len);
		ret = litefs_receive_data(buf,len);
	}
	return ret;
}
static int litefs_utimes (const char *fname, struct timeval *tvp)
{
	int ret = DISPATCH_DONTCARE;
	LITEFS_INFO info;
	LITE_REQUEST_DATA req;
	int len;
	if (litefs_translate (fname,info,req,len)!=-1){
		req.cmd = LITE_UTIMES;
		if (tvp == NULL){
			req.mode = -1;
			req.perm = -1;
		}else{
			req.mode = tvp->tv_sec;
			req.perm = tvp->tv_usec;
		}
		litefs_sendreq (req,len);
		ret = litefs_receive();
	}
	return ret;
}

static int litefs_access (const char *fname, int type)
{
	int ret = DISPATCH_DONTCARE;
	LITEFS_INFO info;
	LITE_REQUEST_DATA req;
	int len;
	if (litefs_translate (fname,info,req,len)!=-1){
		req.cmd = LITE_ACCESS;
		req.mode = type;
		litefs_sendreq (req,len);
		ret = litefs_receive();
	}
	return ret;
}

static DISPATCH m={
	NULL,
	litefs_open,
	litefs_close,
	litefs_write,
	litefs_read,
	NULL,
	litefs_lseek,
	litefs_stat,
	litefs_fstat,
	litefs_lstat,
	litefs_chdir,
	litefs_opendir,
	litefs_readdir,
	litefs_closedir,
	litefs_rename,
	litefs_unlink,
	litefs_rmdir,
	litefs_mkdir,
	NULL,		// socket
	NULL,		// connect
	NULL,		// bind
	litefs_chmod,
	litefs_chown,
	litefs_lchown,
	litefs_readlink,
	litefs_utimes,
	litefs_access,
	litefs_link,
	litefs_symlink,
	litefs_fchmod,
	litefs_fchown,
};

static REGISTER_DISPATCH register_m (m);





