#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <dirent.h>
#include <signal.h>
#include <sys/time.h>
#include <utime.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include "litefs.h"
#include "litefs.p"

static bool debug=false;

static void n_debug (const char *ctl, ...)
{
	if (debug){
		va_list list;
		va_start (list,ctl);
		vfprintf (stderr,ctl,list);
		va_end (list);
	}
}


static void usage()
{
	fprintf (stderr,"litefsd version 1.0\n"
		"\n"
		"litefsd --pipe [ --fork ] [ --debug ] [ --chdir path ]\n"
		"litefsd --socket sock_path [ --fork ] [ --debug ] [ --chdir path ]\n");
}

static const char *workdir;
static int len_workdir;

static void litefsd_2abs(
	const char *arg,
	char abspath[PATH_MAX])
{
	if (arg[0] == '/' && arg[1] == '/'){
		strncpy (abspath,arg,PATH_MAX);
	}else{
		strcpy (abspath,workdir);
		strncpy (abspath+len_workdir,arg,PATH_MAX-len_workdir);
	}
	abspath[PATH_MAX-1] = '\0';
}


/*
	Convert the request to an absolute path by matching a volume
	definition. Send failure to the client if there is no matching
	volume or the request is invalid (.. in path)

	Return -1 if this is not a valid request.
*/
static int litefsd_translate (
	int fd,
	LITE_REQUEST_DATA &req,
	char abspath[PATH_MAX])
{
	int ret = -1;
	n_debug ("translate len = %d\n",req.len);
	if (req.len > 0){
		/* #Specification: litefs server / mapping the working dir
			litefs operates in the current directory unless it is forced
			to operate in the root. The users sees a litefs volume as
			a sub-directory (~/n/volume), so when it access a file
			~/n/volume/file is translate on the litefsd side to
			current_dir/file

			If the user specify a path like this: ~/n/volume//file, then
			this is converted to /file on the server side. This was done
			because it is expect that litefs will generally use to access
			a specific hierarchy on the server side, so by using
			the working directory of litefsd, we get this result.
		*/
		litefsd_2abs (req.buf,abspath);
		ret = 0;
	}else{
		litefsd_sendfail (fd);
	}
	return ret;
}


/*
	Read a request from one client
	Every request must contain a credential, to be added later
*/
static int litefsd_doreq (int fin, int fout)
{
	int ret = -1;
	const int maxdir = 1024;
	static DIR *tbdir[maxdir];
	LITE_REQUEST_DATA req;
	int len = read (fin,&req,sizeof(req));
	if (len >= (int)sizeof(LITE_REQUEST)){
		ret = 0;
		len -= sizeof(LITE_REQUEST);
		while (len < req.len){
			int chunk = read (fin,req.buf+len,sizeof(req.buf)-len);
			if (chunk > 0){
				len += chunk;
			}else{
				ret = -1;
				break;
			}
		}
		if (req.cmd == LITE_VERSION){
			if (req.mode != LITEFS_PROTOCOL_VERSION){
				litefsd_sendfail (fout);
			}else{
				litefsd_sendok(fout);
			}
		}else if (req.cmd == LITE_SHUTDOWN){
			litefsd_sendok(fout);
			ret = -1;
		}else if (req.cmd == LITE_OPEN){
			n_debug ("open :%s: %d %d\n",req.buf,req.mode,req.perm);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath)!=-1){
				int handle = open (abspath,req.mode,req.perm);
				n_debug ("open abspath :%s: handle %d\n",abspath,handle);
				if (handle != -1){
					litefsd_sendfd (fout,handle);
				}else{
					litefsd_senderrno(fout);
				}
			}
		}else if (req.cmd == LITE_CLOSE){
			n_debug ("close %d\n",req.handle);
			if (close(req.handle)!=-1){
				litefsd_sendok(fout);
			}else{
				litefsd_senderrno(fout);
			}
		}else if (req.cmd == LITE_SEEK){
			n_debug ("seek %d %d %d\n",req.handle,req.mode,req.perm);
			off_t off = lseek(req.handle,req.mode,req.perm);
			if (off != (off_t)-1){
				litefsd_sendseek(fout,off);
			}else{
				litefsd_senderrno(fout);
			}
		}else if (req.cmd == LITE_READ){
			n_debug ("read %d %d\n",req.handle,req.mode);
			LITE_REPLY_DATA rep;
			rep.len = read (req.handle,rep.buf,req.mode);
			if (rep.len >= 0){
				rep.status = LITE_STATUS_OK;
				write (fout,&rep,sizeof(LITE_REPLY)+rep.len);
			}else{
				litefsd_senderrno(fout);
			}
		}else if (req.cmd == LITE_WRITE){
			int len = write (req.handle,req.buf,req.len);
			if (len >= 0){
				litefsd_sendok(fout,len);
			}else{
				litefsd_senderrno(fout);
			}
		}else if (req.cmd == LITE_STAT){
			n_debug ("stat :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath)!=-1){
				n_debug ("stat abspath :%s:\n",abspath);
				struct stat st;
				if (stat(abspath,&st)==-1){
					litefsd_senderrno(fout);
				}else{
					litefsd_sendstat (fout,st);
				}
			}
		}else if (req.cmd == LITE_LSTAT){
			n_debug ("lstat :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath)!=-1){
				n_debug ("lstat abspath :%s:\n",abspath);
				struct stat st;
				if (lstat(abspath,&st)==-1){
					litefsd_senderrno(fout);
				}else{
					litefsd_sendstat (fout,st);
				}
			}
		}else if (req.cmd == LITE_FSTAT){
			struct stat st;
			if (fstat(req.handle,&st)==-1){
				n_debug ("fstat %d failed %d\n",req.handle,errno);
				litefsd_senderrno(fout);
			}else{
				litefsd_sendstat (fout,st);
				n_debug ("fstat %d ok\n",req.handle);
			}
		}else if (req.cmd == LITE_UNLINK){
			n_debug ("unlink :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath)!=-1){
				n_debug ("unlink abspath :%s:\n",abspath);
				if (unlink(abspath)!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_RENAME){
			const char *newp = req.buf;
			while (*newp != '\0') newp++;
			newp++;
			n_debug ("rename :%s: :%s:\n",req.buf,newp);
			char old_abspath[PATH_MAX];
			if (litefsd_translate (fout,req,old_abspath)!=-1){
				char new_abspath[PATH_MAX];
				litefsd_2abs (newp,new_abspath);
				n_debug ("rename abspath :%s: :%s:\n",old_abspath,new_abspath);
				if (rename(old_abspath,new_abspath)!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_LINK){
			const char *newp = req.buf;
			while (*newp != '\0') newp++;
			newp++;
			n_debug ("link :%s: :%s:\n",req.buf,newp);
			char old_abspath[PATH_MAX];
			if (litefsd_translate (fout,req,old_abspath)!=-1){
				char new_abspath[PATH_MAX];
				litefsd_2abs (newp,new_abspath);
				n_debug ("link abspath :%s: :%s:\n",old_abspath,new_abspath);
				if (link(old_abspath,new_abspath)!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_SYMLINK){
			const char *oldp = req.buf;
			while (*oldp != '\0') oldp++;
			oldp++;
			n_debug ("link :%s: :%s:\n",oldp,req.buf);
			char new_abspath[PATH_MAX];
			if (litefsd_translate (fout,req,new_abspath)!=-1){
				n_debug ("symlink abspath :%s: :%s:\n",oldp,new_abspath);
				if (symlink(oldp,new_abspath)!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_UTIMES){
			n_debug ("utimes :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath)!=-1){
				n_debug ("utimes abspath :%s: %d %d\n",abspath,req.mode,req.perm);
				int ok = -1;
				if (req.mode == -1 && req.perm == -1){
					ok = utimes (abspath,NULL);
				}else{
					struct timeval tvp;
					tvp.tv_sec = req.mode;
					tvp.tv_usec = req.perm;
					ok = utimes (abspath,&tvp);
				}
				if (ok!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_ACCESS){
			n_debug ("access :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath)!=-1){
				n_debug ("access abspath :%s: %d %d\n",abspath,req.mode,req.perm);
				int ok = access (abspath,req.mode);
				if (ok!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_RMDIR){
			n_debug ("rmdir :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath)!=-1){
				n_debug ("rmdir abspath :%s:\n",abspath);
				if (rmdir(abspath)!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_MKDIR){
			n_debug ("mkdir :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath)!=-1){
				n_debug ("mkdir abspath :%s:\n",abspath);
				if (mkdir(abspath,req.mode)!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_CHDIR){
			// chdir means, may chdir
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath)!=-1){
				n_debug ("chdir abspath :%s:\n",abspath);
				struct stat st;
				if (stat(abspath,&st)!=-1 && S_ISDIR(st.st_mode)){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_CHMOD){
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath)!=-1){
				n_debug ("chmod abspath :%s: %o\n",abspath,req.mode);
				if (chmod (abspath,req.mode)!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_FCHMOD){
			n_debug ("fchmod %d\n",req.handle);
			if (fchmod(req.handle,req.mode)!=-1){
				litefsd_sendok (fout);
			}else{
				litefsd_senderrno (fout);
			}
		}else if (req.cmd == LITE_FCHOWN){
			n_debug ("fchown %d\n",req.handle);
			if (fchown(req.handle,req.mode,req.perm)!=-1){
				litefsd_sendok (fout);
			}else{
				litefsd_senderrno (fout);
			}
		}else if (req.cmd == LITE_CHOWN){
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath)!=-1){
				n_debug ("chown abspath :%s: %d %d\n",abspath,req.mode,req.perm);
				if (chown (abspath,req.mode,req.perm)!=-1){
					litefsd_sendok (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_OPENDIR){
			n_debug ("opendir :%s:\n",req.buf);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath)!=-1){
				n_debug ("opendir abspath :%s:\n",abspath);
				DIR *dir = opendir (abspath);
				if (dir != NULL){
					int i;
					for (i=0; i<maxdir; i++){
						if (tbdir[i] == NULL){
							tbdir[i] = dir;
							litefsd_sendhandle (fout,i);
							break;
						}
					}
					if (i==maxdir) litefsd_sendfail (fout);
				}else{
					litefsd_senderrno (fout);
				}
			}
		}else if (req.cmd == LITE_READDIR){
			n_debug ("readdir %d\n",req.handle);
			DIR *dir = tbdir[req.handle];
			if (dir == NULL){
				litefsd_sendfail (fout);
			}else{
				struct dirent *ent = readdir (dir);
				LITE_REPLY_DATA rep;
				rep.status = LITE_STATUS_OK;
				if (ent == NULL){
					rep.len = 0;
				}else{
					rep.len = sizeof(struct dirent);
					memcpy (rep.buf,ent,sizeof(struct dirent));
				}
				write (fout,&rep,sizeof(LITE_REPLY)+rep.len);
			}
		}else if (req.cmd == LITE_CLOSEDIR){
			n_debug ("closedir %d\n",req.handle);
			DIR *dir = tbdir[req.handle];
			if (dir == NULL){
				litefsd_sendfail (fout);
			}else if (closedir(dir)!=-1){
				litefsd_sendok(fout);
			}else{
				litefsd_senderrno(fout);
			}
		}else if (req.cmd == LITE_READLINK){
			n_debug ("readlink %d\n",req.handle);
			char abspath[PATH_MAX];
			if (litefsd_translate (fout,req,abspath)!=-1){
				n_debug ("readlink abspath :%s:\n",abspath);
				LITE_REPLY_DATA rep;
				rep.len = readlink (abspath,rep.buf,sizeof(rep.buf));
				if (len ==-1){
					litefsd_senderrno (fout);
				}else{
					rep.status = LITE_STATUS_OK;
					write (fout,&rep,sizeof(LITE_REPLY)+rep.len);
				}
			}
		}else{
			litefsd_sendfail(fout);
			ret = -1;
		}
	}
	return ret;
}
		



/*
	Process all request and connections. Infinite loop
*/
static void litefsd_loop (
	int sock,		// Unix socket to receive new connection
					// or -1 if not used
	int pipein,		// Pre-connected client or proxy
	int pipeout)	// Pre-connected client or proxy
{
	int tbfd[1024];
	int nbfd = 0;
	fd_set fdin;
	FD_ZERO (&fdin);
	int maxfd = 0;
	if (sock != -1){
		FD_SET (sock,&fdin);
		maxfd = sock + 1;
	}
	if (pipein != -1){
		FD_SET (pipein,&fdin);
		if (pipein >= maxfd) maxfd = pipein+1;
	}
	while (1){
		fd_set in;
		memcpy (&in,&fdin,sizeof(in));
		n_debug ("select maxfd %d\n",maxfd);
		if (select (maxfd,&in,NULL,NULL,NULL)!=-1){
			int orig_nbfd = nbfd;
			if (sock != -1 && FD_ISSET (sock,&in)){
				struct sockaddr_un unc;
				size_t len = sizeof(unc);
				unc.sun_family = AF_UNIX;
				int s = accept (sock,(struct sockaddr*)&unc,&len);
				tbfd[nbfd++] = s;
				FD_SET (s,&fdin);
				if (s >= maxfd) maxfd = s + 1;
			}
			if (pipein != -1 && FD_ISSET(pipein,&in)){
				if (litefsd_doreq (pipein,pipeout) == -1){
					break;
				}
			}
			int *ptfd = tbfd;
			bool redo_fdin = false;
			for (int i=0; i< orig_nbfd; i++, ptfd++){
				int fd = *ptfd;
				if (FD_ISSET(fd,&in)){
					if (litefsd_doreq (fd,fd) == -1){
						close (fd);
						redo_fdin = true;
						*ptfd = -1;
					}
				}
			}
			if (redo_fdin){
				FD_ZERO (&fdin);
				maxfd = 0;
				if (sock != -1){
					maxfd = sock+1;
					FD_SET (sock,&fdin);
				}
				if (pipein != -1){
					FD_SET (pipein,&fdin);
					if (pipein >= maxfd) maxfd = pipein+1;
				}
				ptfd = tbfd;
				int *copyfd = tbfd;
				for (int i=0; i<nbfd; i++, ptfd++){
					int fd = *ptfd;
					if (fd != -1){
						FD_SET (fd,&fdin);
						if (fd >= maxfd) maxfd = fd + 1;
						*copyfd++ = fd;
					}
				}
				nbfd = (int)(copyfd - tbfd);
			}
		}
	}
}

int main (int argc, char *argv[])
{
	int ret = -1;
	if (argc == 1){
		usage();
	}else{
		int sock = -1;
		int pipein = -1;
		int pipeout = -1;
		const char *newdir = NULL;
		bool dofork = false;
		for (int i=1; i<argc; i++){
			if (strcmp(argv[i],"--socket")==0){
				i++;
				sock = litefsd_server(argv[i]);
			}else if (strcmp(argv[i],"--pipe")==0){
				pipein = 0;
				pipeout = 1;
			}else if (strcmp(argv[i],"--chdir")==0){
				i++;
				newdir = argv[i];
			}else if (strcmp(argv[i],"--debug")==0){
				debug = true;
			}else if (strcmp(argv[i],"--fork")==0){
				dofork = true;
			}else{
				usage();
				exit (-1);
			}
		}
		if (newdir != NULL && chdir(newdir)==-1){
			fprintf (stderr,"Can't change to directory %s (%s)\n"
				,newdir,strerror(errno));
			exit (-1);
		}
		char cwd[PATH_MAX];
		if (getcwd(cwd,PATH_MAX)==NULL){
			fprintf (stderr,"Can't locate current working directory (%s)\n"
				,strerror(errno));
		}else{
			workdir = cwd;
			len_workdir = strlen(cwd);
			if (sock != -1 || pipein != -1){
				if (dofork){
					if (fork() != 0) exit(0);
					setsid();
				}
				litefsd_loop(sock,pipein,pipeout);
			}else{
				usage();
			}
		}
	}
	return ret;
}



